use crate::{
    app::session::SessionToken,
    util::pack::pack_u16,
};

use super::Packet;

#[derive(Clone)]
pub struct PacketGameState {
    pub token:SessionToken,
}
impl PacketGameState {
    pub fn new() -> Self
    {
        Self {
            token:SessionToken::default(),
        }
    }
}
impl Packet for PacketGameState {
    type Data = Self;

    fn decode(data:&Vec<u8>, index:&mut usize) -> Result<Self::Data, ()>
    {
        let mut result = Self::new();

        if data.len() - *index == 8 {
            for i in 0..8 {
                result.token[i] = data[*index];
                *index += 1;
            }
            Ok(result)
        } else {
            Err(())
        }
    }
}


#[derive(Clone, Copy)]
pub struct PacketGameStateResponsePiece {
    pub valid:bool,
    pub piece:u8,
    pub promoted:bool,
    pub player:u8,
    pub tile:u8,
}
impl PacketGameStateResponsePiece {
    pub fn new() -> Self
    {
        Self {
            valid:false,
            piece:0,
            promoted:false,
            player:0,
            tile:0,
        }
    }
}

#[derive(Clone)]
pub struct PacketGameStateResponse {
    pub status:u16,
    pub turn:u16,
    pub dawn_handle:String,
    pub dusk_handle:String,
    pub dawn_pool:[u8; 6],
    pub dusk_pool:[u8; 6],
    pub pieces:[PacketGameStateResponsePiece; 38],
}
impl PacketGameStateResponse {
    pub fn new() -> Self
    {
        Self {
            status:0,
            turn:0,
            dawn_handle:String::new(),
            dusk_handle:String::new(),
            dawn_pool:[0; 6],
            dusk_pool:[0; 6],
            pieces:[PacketGameStateResponsePiece::new(); 38],
        }
    }
}
impl Packet for PacketGameStateResponse {
    type Data = Self;

    fn encode(&self) -> Vec<u8>
    {
        let mut piece_bytes = Vec::new();
        for piece in &self.pieces {
            let piece_data: u16 = piece.valid as u16
            | ((piece.piece as u16) << 1)
            | ((piece.promoted as u16) << 4)
            | ((piece.player as u16) << 5)
            | ((piece.tile as u16) << 6);

            piece_bytes.append(&mut pack_u16(piece_data));
        }

        let dawn_pool_bytes :u16 = self.dawn_pool[0] as u16
        | ((self.dawn_pool[1] as u16) << 5)
        | ((self.dawn_pool[2] as u16) << 7)
        | ((self.dawn_pool[3] as u16) << 9)
        | ((self.dawn_pool[4] as u16) << 11)
        | ((self.dawn_pool[5] as u16) << 13);

        let dusk_pool_bytes :u16 = self.dusk_pool[0] as u16
        | ((self.dusk_pool[1] as u16) << 5)
        | ((self.dusk_pool[2] as u16) << 7)
        | ((self.dusk_pool[3] as u16) << 9)
        | ((self.dusk_pool[4] as u16) << 11)
        | ((self.dusk_pool[5] as u16) << 13);

        [
            pack_u16(self.status),
            pack_u16(self.turn),
            pack_u16(self.dawn_handle.len() as u16),
            self.dawn_handle.as_bytes().to_vec(),
            pack_u16(self.dusk_handle.len() as u16),
            self.dusk_handle.as_bytes().to_vec(),
            pack_u16(dawn_pool_bytes),
            pack_u16(dusk_pool_bytes),
            piece_bytes,
        ].concat()
    }
}
